; implement the given buffer in the file
; written for DATAdesign engine v2.+
; (c)1992 by Joachim Van der Auwera for PROGS
        section text
        include aid_mac
        include struct_h
        include dev8_err
*
; implement the given buffer in the file
; if file based then record will be written to the file for complete
; safety in case of a possible system crash or power failure
; indexes are not adjusted to the changes !!
; can cause problems: file errors, write protected disk
; solved: DRIVE FULL, a record is not (re-)implemented when there is danger for a drive full error
; parameters
p.id    equ     $4	    ; buffer id (long word, optional)
; errors
; rdo   sorry, this is a read only buffer
; imem  out of memory
; drfl  drive full
;       too many records, impossible errors as a 32 bit compater can't have enough memory to store the retrieval index
; any file i/o error (should only be drfl)
cod_IMPL xsub
        regsave d1-d3/d6/d7/a0-a5
        move.l  a1,a5	  ; keep params in a safe place

        move.l  p.id(a5),d0	  ; buffer id if passed
        tst.w   p.id-4(a5)	  ; is the buffer id given ?
        bne.s   tbuff	  ; yes -> ..
        xjsr    getbuff	  ; get default buffer
tbuff   move.l  d0,a4	  ; keep it
        xjsr    testbid	  ; and test if it is valid
        bne     codend	  ; if not -> error

        moveq   #err.rdo,d0	  ; assume we can't write on this buffer
        tst.b   b_view(a4)	  ; check it out
        bne     codend	  ; you see ...

        moveq   #0,d0
        tst.b   b_change(a4)	  ; do we have to reimplement this one ?
        beq     codend	  ; no -> exit without error

        tst.l   b_rlen(a4)	  ; is there something to implement
        beq     codend	  ; no -> exit without error

        move.l  b_file(a4),a3   ; get fileid

        tst.b   f_file(a3)	  ; file based ??
        bne     implfile	  ; so implement it on file (what else)

        tst.b   b_new(a4)	  ; is this a new record
        bne.s   newrec	  ; yes it is -> ...

        ; re-implement the record
        mutex		  ; hey, we don't want no mixup
        move.l  b_id(a4),d1	  ; recordid
        xjsr    findrec	  ; get the record
        bne.s   ahum	  ; ahum, what's going on, this is IMPOSSIBLE

        ; can we fit it in ??
        move.l  rl_pos(a1),a2   ; ptr to record in memory
        move.l  b_rlen(a4),d1   ; record length
        cmp.l   r_space(a2),d1  ; does it fit ?
        ble.s   fits	  ; yahoe, now just copy the lot, and fill in the new length

        ; so get the new buffer memory
        add.l   #r_fld,d1	  ; add header length
        xjsr    getsmem	  ; get the memory for it
        bne.s   ahum	  ; so, we're out of memory are we ??

        sub.l   #r_fld,d1	  ; length of reserved space without header length
        move.l  d1,r_space(a0)  ; length of reserved space
        move.l  b_id(a4),r_id(a0);fill in recordid

        move.l  rl_pos(a1),a2   ; start of area to release
        move.l  a0,rl_pos(a1)   ; new record ptr
        exg     a0,a2	  ; exchange old and new
        xjsr    relmem	  ; release it

fits    ; now copy the record and fill in the length (a2)
        move.l  b_rlen(a4),d1   ; record length
        move.l  d1,r_len(a2)	  ; fill it in
        move.b  b_mark(a4),r_mark(a2)	; set mark status
        xjsr    getdate	  ; and also date of last adjustment
        move.l  d0,r_date(a2)   ; fillin
        lsr.l   #1,d1	  ; nr of words to copy
        move.l  b_mem(a4),a0	  ; start of buffer to copy
        adda.w  #r_fld,a2	  ; skip header
copy    move.w  (a0)+,(a2)+	  ; copy a word
        subq.l  #1,d1	  ; another one ?
        bgt     copy	  ; yes -> ..

        ; and now this record isn't changed any more
        clr.b   b_change(a4)	  ; so clear the status
        moveq   #0,d0
ahum
        endmutex
codend
        subend

newrec  ; we have to implement a new record
        mutex

        ; try to find a suitable recordid
        move.l  f_lstrid(a3),d3 ; previous id
nextid  addq.l  #1,f_lstrid(a3) ; try next one
        move.l  f_lstrid(a3),d1 ; get it
        cmpi.l  #-1,d1	  ; test for the impossible recordid
        beq.s   nextid	  ; oops, take next one
        moveq   #err.drfl,d0	  ; assume we can't add another record
        cmp.l   d1,d3	  ; check it out
        beq     ahum	  ; no, you can't mean this, you have more than 4e9 record ???
			  ; this can't even fit 32bit memory
        xjsr    findrec	  ; check if it is used already
        beq.s   nextid	  ; it is, so try again
        ; f_lstrid now contains a suitable recordid

        ; we now need an empty record linkage
        move.l  f_free(a3),d1   ; is there one ?
        bne.s   gotfree	  ; yes -> ...

        ; allocate some memory for 100 new record linkages
        move.l  #rl.100,d1	  ; amount of mem for it
        xjsr    getsmem	  ; get the memory
        bne     ahum	  ; imem -> ..

        move.l  f_area(a3),(a0) ; link the area in
        move.l  a0,f_area(a3)
        adda.w  #4,a0

        moveq   #99,d1	  ; nr of links
lnkloop move.l  f_free(a3),(a0) ; link in
        move.l  a0,f_free(a3)
        move.l  #-1,rl_id(a0)   ; and mark as unused
        adda.w  #rl.len,a0	  ; go to next one
        dbf     d1,lnkloop	  ; and link next one

        ; now use one of these links
gotfree
        ; so get the new buffer memory
        move.l  b_rlen(a4),d1   ; get record length
        add.l   #r_fld,d1	  ; add header length
        xjsr    getsmem	  ; get the memory for it
        bne     ahum	  ; so we're out of memory are we ??

        sub.l   #r_fld,d1	  ; length of reserved space without header length
        move.l  d1,r_space(a0)  ; length of reserved space
        move.l  f_lstrid(a3),r_id(a0)	; fill in recordid

        move.l  f_free(a3),a2   ; get the linkage
        move.l  (a2),f_free(a3) ; unlink it from the free-list
        ; now fill it in (partially)
        move.l  f_lstrid(a3),rl_id(a2)	; fill in recordid
        move.l  a0,rl_pos(a2)   ; fill in ptr to record
        move.b  #1,rl_edit(a2)  ; this record is being edited

        ; this record isn't new any more
        move.l  f_lstrid(a3),b_id(a4)	; fill in recordid
        clr.b   b_new(a4)	  ; not new any more

        ; now we have to link it and fill in the record (preserve a2 during linking)
        ; linkin is recursive routine to link a record in
        ; ptr to list: on stack, a2 linkage block, d2 boolean
        lea     f_top(a3),a0
        move.l  a0,-(sp)	  ; head pointer of tree
        moveq   #0,d2	  ; start with false
        jsr     linkin
        adda.w  #4,sp	  ; clean up stack

        ; and we have an extra record in the file
        addq.l  #1,f_count(a3)

        ; fill in the record (recordlinkage in a2)
        move.l  rl_pos(a2),a2   ; start of actual record
        bra     fits

; linkin is recursive routine to link a record in
; addr of ptr to list: on stack, a2 linkage block (preserved), d2 boolean
; based on zoek routine in "Algoritmen en datastructuren" p284
; can corrupt a0/a1/a5 and d1
        xdef    linkin
p.ptr   equ     $4		  ; place, relative to stack where we can find the head ptr of the tree
linkin
        move.l  p.ptr(sp),a0	  ; ptr to head ptr
        tst.l   (a0)	  ; zero ?
        bne.s   balance	  ; no -> figure out a balanced tree
        move.l  a2,(a0)	  ; fill in ptr
        clr.l   rl_left(a2)
        clr.l   rl_right(a2)
        clr.b   rl_avl(a2)
        moveq   #1,d2	  ; branche has grown
elink   rts		  ; and that's it
balance
        move.l  (a0),a0	  ; current linkage
        move.l  rl_id(a2),d1	  ; id of new record
        cmp.l   rl_id(a0),d1	  ; compare with top of tree
        bpl     right	  ; continue with right branche
        lea     rl_left(a0),a0
        move.l  a0,-(sp)	  ; head pointer of tree (left branche)
        jsr     linkin	  ; !! no register values are preserved !! (recursive call)
        adda.w  #4,sp	  ; clean up stack
        move.l  p.ptr(sp),a0	  ; and restore pointer
        move.l  (a0),a0	  ; to value as at the start
        tst.b   d2		  ; has there been any growth ?
        beq     elink	  ; no -> in balance
        tst.b   rl_avl(a0)	  ; how WAS the balance
        beq.s   lcase0	  ; when was zero
        bmi.s   lcase_1	  ;	   -1
lcase1  clr.b   rl_avl(a0)	  ; balance is perfect
        moveq   #0,d2	  ; no growth
        rts
lcase0  move.b  #-1,rl_avl(a0)  ; set balance
        rts
lcase_1 move.l  rl_left(a0),a1  ; set p1
        tst.b   rl_avl(a1)	  ; check balance
        bpl.s   l_else	  ; <> -1 -> ..
        ; enkelvoudig LL roteren
        move.l  rl_right(a1),rl_left(a0); p^.links=p1^.rechts
        move.l  a0,rl_right(a1) ; p1^rechts=p
        clr.b   rl_avl(a0)	  ; clear balance
        move.l  p.ptr(sp),a0
        move.l  a1,(a0)	  ; p=p1
        clr.b   rl_avl(a1)	  ; also clear p^bal(=p1^bal)
        moveq   #0,d2	  ; no growth
        rts
l_else  ; dubbel LR roteren
        move.l  a2,a5	  ; keep a2 for a while
        move.l  rl_right(a1),a2 ; p2=p1^rechts
        move.l  rl_left(a2),rl_right(a1)
        move.l  a1,rl_left(a2)
        move.l  rl_right(a2),rl_left(a0)
        move.l  a0,rl_right(a2)
        clr.b   rl_avl(a0)	  ; get balances figured out
        tst.b   rl_avl(a2)
        bpl.s   LR_1
        move.b  #1,rl_avl(a0)
LR_1    clr.b   rl_avl(a1)
        tst.b   rl_avl(a2)
        ble.s   LR_2
        move.b  #-1,rl_avl(a0)
LR_2
        move.l  p.ptr(sp),a1	  ; get value for p
        move.l  a2,(a1)	  ; p=p2
        clr.b   rl_avl(a2)	  ; in balance p^bal=0 (p aangepast)
        moveq   #0,d2	  ; no growth
        move.l  a5,a2	  ; restore a2
        rts

right   lea     rl_right(a0),a0
        move.l  a0,-(sp)	  ; head pointer of tree (right branche)
        jsr     linkin	  ; !! no register values are preserved !! (recursive call)
        adda.w  #4,sp	  ; clean up stack
        move.l  p.ptr(sp),a0	  ; and restore pointer
        move.l  (a0),a0	  ; to value as at the start
        tst.b   d2		  ; has there been any growth ?
        beq     elink	  ; no -> in balance
        tst.b   rl_avl(a0)	  ; how WAS the balance
        bmi     lcase1	  ; when was -1 , is same as 1 for other case
        bgt.s   rcase1	  ; when was 1
rcase0  move.b  #1,rl_avl(a0)   ; set balance
        rts
rcase1  move.l  rl_right(a0),a1 ; set p1
        tst.b   rl_avl(a1)	  ; check balance
        ble.s   r_else	  ; <> 1 -> ..
        ; enkelvoudig RR roteren
        move.l  rl_left(a1),rl_right(a0); p^.rechts=p1^.links
        move.l  a0,rl_left(a1)  ; p1^links=p
        clr.b   rl_avl(a0)	  ; clear balance
        move.l  p.ptr(sp),a0	  ; get value for p
        move.l  a1,(a0)	  ; p=p1
        clr.b   rl_avl(a1)	  ; also clear p^bal(=p1^bal)
        moveq   #0,d2	  ; no growth
        rts
r_else  ; dubbel RL roteren
        move.l  a2,a5	  ; keep a2 for a while
        move.l  rl_left(a1),a2  ; p2=p1^links
        move.l  rl_right(a2),rl_left(a1)
        move.l  a1,rl_right(a2)
        move.l  rl_left(a2),rl_right(a0)
        move.l  a0,rl_left(a2)
        clr.b   rl_avl(a0)	  ; get balances figured out
        tst.b   rl_avl(a2)
        ble.s   RL_1
        move.b  #-1,rl_avl(a0)
RL_1    clr.b   rl_avl(a1)
        tst.b   rl_avl(a2)
        bpl.s   RL_2
        move.b  #1,rl_avl(a0)
RL_2
        move.l  p.ptr(sp),a1	  ; get value for p
        move.l  a2,(a1)	  ; p=p2
        clr.b   rl_avl(a2)	  ; in balance p^bal=0 (p=p2)
        moveq   #0,d2	  ; no growth
        move.l  a5,a2	  ; restore a2
        rts
;
eof     equ     $3fffffff	  ; file pos to move to end of file (max file len 1 GIGABYTE)
implfile; this file is stored on a medium and has to implemented there
        ; a3/a4 are set, we are is user mode !!
        ; ahum goes back to user and ends routine
        mutex		  ; hey, we don't want no mixup
        tst.b   b_new(a4)	  ; is this a new record
        bne     f_new	  ; yes it is -> ...

        ; just check if we can write to this medium
        move.l  f_chid(a3),a0   ; channel id
        xjsr    iof_frsp	  ; get nr of free bytes in d1
        bne     ahum	  ; oops, read only

        ; re-implement the record
        move.l  b_id(a4),d1	  ; recordid
        xjsr    findrec	  ; get the record
        bne     ahum	  ; ahum, what's going on, this is IMPOSSIBLE
        move.l  a1,a5	  ; keep it for a while

        ; can we fit it in ??
        ; get recordlen first (d2)
        move.l  f_chid(a3),a0   ; channel id
        move.l  rl_pos(a5),d1   ; file position
        addi.l  #10,d1	  ; skip 'R'.b,mark.b,id.l,date.l
        xjsr    iof_posa
        bne     ahum	  ; ahum...

        ; read record length
        moveq   #0,d2	  ; clear record len
        xjsr    iob_fbyt
        bne     ahum
        move.b  d1,d2
        xjsr    iob_fbyt
        bne     ahum
        lsl.l   #8,d2
        move.b  d1,d2
        xjsr    iob_fbyt
        bne     ahum
        lsl.l   #8,d2
        move.b  d1,d2
        xjsr    iob_fbyt
        bne     ahum
        lsl.l   #8,d2
        move.b  d1,d2
        ; so old record len in in d2
        cmp.l   b_rlen(a4),d2   ; compare
        beq.s   savest	  ; same -> save record at old position
        bmi     testsp	  ; new longer -> see where we have to fit it

        ; so the new record is shorter
        ; we first have to clear the end of the record
        ; go to start of area to clear
        sub.l   b_rlen(a4),d2   ; nr of bytes we have to clear
        move.l  rl_pos(a5),d1   ; start of record
        add.l   #14,d1	  ; skip record data
        add.l   b_rlen(a4),d1   ; and end of new record
        xjsr    iof_posa	  ; position file pointer
        bne     ahum	  ; oops
        ; and clear it
        moveq   #0,d1	  ; byte to send
f_cllp
        xjsr    iob_sbyt	  ; send a byte
        bne     ahum	  ; oops
        subq.l  #1,d2	  ; another one ?
        bgt.s   f_cllp

savest  ; now save record at start of record
        move.l  rl_pos(a5),d1   ; position
        xjsr    iof_posa
        bne     ahum	  ; oops
        bsr.s   saverec	  ; save the record
        bne     ahum	  ; oops

flush   ; and now this record isn't changed any more
        clr.b   b_change(a4)	  ; so clear the status
        endmutex
        xjsr    iof_flsh	  ; flush buffers
        bra     codend	  ; done

; save a record to file
; needs a0/a3/a4 to be set, a5 is record linkage
; has to be called from supervisor
saverec suba.w  #20,sp	  ; we need some stackspace
        move.l  sp,a1	  ; start of buffer
        move.l  rl_pos(a4),a2   ; actual record
        move.b  #'R',(a1)+	  ; byte to send
        move.b  b_mark(a4),(a1)+
        move.l  rl_id(a5),(a1)+
        xjsr    getdate
        move.l  d0,(a1)+
        move.l  b_rlen(a4),(a1)+
        move.l  sp,a1	  ; start of buffer
        moveq   #14,d2	  ; nr of bytes to send
        xjsr    iob_smul
        bmi.s   oops
        ; now the record contents
        move.l  b_rlen(a4),d2   ; record len
        move.l  b_mem(a4),a1	  ; buffer start
        xjsr    iob_smul
        bne.s   oops
        ; now adjust lstrid on disk
        moveq   #18,d1	  ; filepos
        xjsr    iof_posa
        bne.s   oops
        move.l  sp,a1	  ; send it
        move.l  f_lstrid(a3),(a1)
        moveq   #4,d2
        xjsr    iob_smul
oops    adda.w  #20,sp
        rts

f_new   ; we have to implement a new record
        ; first see if there is enough space on disk
        move.l  f_chid(a3),a0   ; channel id
        xjsr    iof_frsp	  ; get nr of free bytes in d1
        bne     ahum	  ; oops, read only
        moveq   #err.drfl,d0	  ; assume there may not be enough space left
        moveq   #14,d2	  ; nr of bytes we need for record header
        add.w   f_irsp(a3),d2   ; and IRS
        add.l   b_rlen(a4),d2   ; and record itself
        cmp.l   d1,d2	  ; compare
        bgt     ahum	  ; oops, plausible drive full

        ; try to find a suitable recordid
        move.l  f_lstrid(a3),d3 ; previous id
fnxtid  addq.l  #1,f_lstrid(a3) ; try next one
        move.l  f_lstrid(a3),d1 ; get it
        cmpi.l  #-1,d1	  ; test for the impossible recordid
        beq.s   fnxtid	  ; oops, take next one
        moveq   #err.drfl,d0	  ; assume we can't add another record
        cmp.l   d1,d3	  ; check it out
        beq     ahum	  ; no, you can't mean this, you have more than 4e9 record ???
			  ; this can't even fit 32bit memory
        xjsr    findrec	  ; check if it is used already
        beq.s   fnxtid	  ; it is, so try again
        ; f_lstrid now contains a suitable recordid

        ; we now need an empty record linkage
        move.l  f_free(a3),d1   ; is there one ?
        bne.s   fgotfree	  ; yes -> ...

        ; allocate some memory for 100 new record linkages
        move.l  #rl.100,d1	  ; amount of mem for it
        xjsr    getsmem	  ; get the memory
        bne     ahum	  ; imem -> ..

        move.l  f_area(a3),(a0) ; link the area in
        move.l  a0,f_area(a3)
        adda.w  #4,a0

        moveq   #99,d1	  ; nr of links
flnklp  move.l  f_free(a3),(a0) ; link in
        move.l  a0,f_free(a3)
        move.l  #-1,rl_id(a0)   ; and mark as unused
        adda.w  #rl.len,a0	  ; go to next one
        dbf     d1,flnklp	  ; and link next one

        ; now use one of these links
fgotfree
        ; position the file pointer
        move.l  f_chid(a3),a0   ; channel id
        move.l  #eof,d1	  ; at end of file
        xjsr    iof_posa	  ; position file pointer
        move.l  d1,d6	  ; keep current filepos for a while

        ; save the IRS
        moveq   #0,d3
        move.w  f_irsp(a3),d3   ; save inter record space
        add.l   d3,d6	  ; adjust file pointer
        subq.w  #1,d3	  ; prepare for dbf
        bmi.s   endirs	  ; if there is some irs
irslp   moveq   #0,d1	  ; bytes to send
        xjsr    iob_sbyt
        bmi     ahum
        dbf     d3,irslp
endirs
        move.l  f_free(a3),a5   ; get the linkage
        move.l  f_lstrid(a3),rl_id(a5)	; fill in recordid
        bsr     saverec	  ; save the record
        move.l  #-1,rl_id(a5)   ; in case of error
        tst.l   d0
        bne     ahum	  ; oops

        move.l  f_free(a3),a5   ; get the linkage
        move.l  (a5),f_free(a3) ; unlink it from the free-list
        ; now fill it in (partially)
        move.l  f_lstrid(a3),rl_id(a5)	; fill in recordid
        move.l  d6,rl_pos(a5)   ; fill in ptr to record
        move.b  #1,rl_edit(a5)  ; this record is being edited

        ; this record isn't new any more
        move.l  f_lstrid(a3),b_id(a4)	; fill in recordid
        clr.b   b_new(a4)	  ; not new any more

        ; now we have to link it and fill in the record (preserve a2 during linking)
        ; linkin is recursive routine to link a record in
        ; ptr to list: on stack, a2 linkage block, d2 boolean
        move.l  a5,a2	  ; linkage
        lea     f_top(a3),a0
        move.l  a0,-(sp)	  ; head pointer of tree
        moveq   #0,d2	  ; start with false
        jsr     linkin
        adda.w  #4,sp	  ; clean up stack

        ; and we have an extra record in the file
        addq.l  #1,f_count(a3)
        ; and now this record isn't changed any more
        clr.b   b_change(a4)	  ; so clear the status
        endmutex
        move.l  f_chid(a3),a0
        xjsr    iof_flsh
        bra     codend	  ; done

; new record is longer than old one
testsp  move.l  b_rlen(a4),d3
        sub.l   d2,d3	  ; nr of empty bytes we need
        move.l  rl_pos(a5),d1
        add.l   d2,d1
        addi.l  #14,d1	  ; end of record
        xjsr    iof_posa	  ; position file pointer
        bne.s   saveend	  ; oops, position at end of file
tstlp
        xjsr    iob_fbyt	  ; read a byte
        bne.s   saveend	  ; problem, save at end
        cmpi.b  #'R',d1	  ; start of next record ?
        beq.s   saveend	  ; yes -> save at end
        tst.l   d3
        subq.l  #1,d3	  ; one less byte to go
        bgt.s   tstlp
        bra     savest
saveend
        ; first see if there is enough space on disk
        xjsr    iof_frsp	  ; get nr of free sectors
        bne     ahum	  ; oops, read only
        moveq   #err.drfl,d0	  ; assume there may be a drive full error
        moveq   #14,d3	  ; nr of bytes we need for record header
        add.w   f_irsp(a3),d3   ; and IRS
        add.l   b_rlen(a4),d3   ; and record itself
        cmp.l   d1,d3	  ; compare
        bgt     ahum	  ; oops, plausible drive full

        move.l  rl_pos(a5),d1   ; position
        addi.l  #14,d1	  ; start of record contents
        xjsr    iof_posa
        bne     ahum	  ; oops
        moveq   #0,d1
rclrlp
        xjsr    iob_sbyt	  ; clear a byte
        bne     ahum
        subq.l  #1,d2	  ; another one ??
        bgt.s   rclrlp

        move.l  rl_pos(a5),d1   ; position
        xjsr    iof_posa
        bne     ahum	  ; oops
        moveq   #14,d2	  ; clear record header, nr of bytes
        moveq   #0,d1
rclrlp2
        xjsr    iob_sbyt	  ; clear a byte
        bne     ahum
        subq.l  #1,d2	  ; another one ??
        bgt.s   rclrlp2

        ; position the file pointer
        move.l  #eof,d1	  ; at end of file
        xjsr    iof_posa	  ; position file pointer
        move.l  d1,d6	  ; keep current filepos for a while

        ; save the IRS
        moveq   #0,d3
        move.w  f_irsp(a3),d3   ; save inter record space
        add.l   d3,d6	  ; adjust file pointer
        subq.w  #1,d3	  ; prepare for dbf
        bmi.s   fendirs	   ; if there is some irs
firslp  moveq   #0,d1	  ; bytes to send
        xjsr    iob_sbyt
        bmi     ahum
        dbf     d3,firslp
fendirs
        move.l  d6,rl_pos(a5)
        bsr     saverec	  ; save the record
        bne     ahum	  ; oops
        bra     flush
;
        end
