; aid routines for file access on a medium from supervisor mode
; written for DATAdesign engine v2.+
; (c)1992 by Joachim Van der Auwera for PROGS
        section text
        include aid_mac
        include dev8_err
        include dev8_qdos_io
*
; send multiple bytes to a file
; parameters :  d2      nr of bytes to send
;               a0      channel id
;               a1      base of buffer
; return        d0      error code
; errors
; drfl  drive full
; ichn  invalid channel id
iob_smul xsub
        regsave d1-d3/a1
        moveq   #forever,d3     ; no timeout
agai1   moveq   #0,d1           ; nr of bytes received
        moveq   #iob.smul,d0    ; send multiple bytes
        trap    #do.io
        cmpi.l  #err.drfl,d0
        beq.s   stop1
        andi.l  #$0000ffff,d1   ; clear the high word
        sub.l   d1,d2           ; any bytes left
        bne.s   agai1
stop1   subend
*
; send a byte to a file
; parameters :  d1      byte to send
;               a0      channel id
; return        d0      error code
; errors
; drfl  drive full
; ichn  invalid channel id
iob_sbyt xsub
        regsave d1-d3/a1
        moveq   #forever,d3     ; no timeout
        move.b  d1,d2
agai2   move.b  d2,d1
        moveq   #iob.sbyt,d0    ; send a byte
        trap    #do.io
        cmpi.l  #err.nc,d0      ; not complete ?
        beq.s   agai2           ; continue
        subend
*
; fetch multiple bytes to a file
; parameters :  d2      nr of bytes to fetch
;               a0      channel id
;               a1      base of buffer
; return        d0      error code
; errors
; eof   end of file
; ichn  invalid channel id
iob_fmul xsub
        regsave d1-d3/a1
        moveq   #forever,d3     ; no timeout
agai3   moveq   #0,d1           ; nr of bytes received
        moveq   #iob.fmul,d0    ; fetch multiple bytes
        trap    #do.io
        cmpi.l  #err.eof,d0
        beq.s   stop3
        andi.l  #$0000ffff,d1   ; clear the high word
        sub.l   d1,d2           ; any bytes left
        bne.s   agai3
stop3   subend
; same but from user mode
iou_fmul xsub
        regsave d1/d3/a1
        moveq   #forever,d3     ; no timeout
agai7   moveq   #0,d1           ; nr of bytes received
        moveq   #iob.fmul,d0    ; fetch multiple bytes
        trap    #do.io
        cmpi.l  #err.eof,d0
        beq.s   stop7
        andi.l  #$0000ffff,d1   ; clear the high word
        sub.l   d1,d2           ; any bytes left
        bne.s   agai7
stop7   subend
*
; fetch a byte from a file
; parameters :  a0      channel id
; return        d0      error code
;               d1      byte fetched
; errors
; drfl  drive full
; ichn  invalid channel id
iob_fbyt xsub
        regsave d2/d3/a1
        moveq   #forever,d3     ; no timeout
        move.b  d1,d2
agai4   move.b  d2,d1
        moveq   #iob.fbyt,d0    ; get a byte
        trap    #do.io
        cmpi.l  #err.nc,d0      ; not complete ?
        beq.s   agai4           ; continue
        subend
*
; position file absolute
; parameters :  d1      new file position
;               a0      channel id
; return        d0      error code
; errors
; ichn  invalid channel id
; eof   end of file
iof_posa xsub
        regsave d3/a1
        moveq   #forever,d3     ; no timeout
agai5   moveq   #iof.posa,d0    ; position file absolute
        trap    #do.io
        cmpi.l  #err.nc,d0      ; not complete ?
        beq.s   agai5           ; continue
        subend
*
; position file relative
; parameters :  d1      displacement in file
;               a0      channel id
; return        d0      error code
; errors
; ichn  invalid channel id
; eof   end of file
iof_posr xsub
        regsave d3/a1
        moveq   #forever,d3     ; no timeout
agai6   moveq   #iof.posr,d0    ; position file relative
        trap    #do.io
        cmpi.l  #err.nc,d0      ; not complete ?
        beq.s   agai6           ; continue
        subend
*
; flush buffers
; parameters :  a0      channel id
; return        d0      error code
; errors
; ichn  invalid channel id
; NOT ATOMIC, don't call from supervisor !!
iof_flsh xsub
        regsave d1/d3/a1
        moveq   #forever,d3     ; no timeout
        moveq   #iof.flsh,d0    ; flush buffers
        trap    #do.io
        subend
*
; get nr of free bytes on medium
; parameters :  a0      channel id
; return        d0      error code
;               d1.l    nr of free bytes
; errors
; ichn  invalid channel id
iof_frsp xsub
        regsave d2-d3/a1-a3
        getssp  ioi.blkl
        move.l  #-1,d2          ; jobid (do we need it?)
        moveq   #forever,d3     ; no timeout
agai9   move.l  sp,a1           ; place for medium information
        moveq   #iof.xinf,d0    ; get medium information
        trap    #do.io
        cmpi.l  #err.nc,d0      ; not complete ?
        beq.s   agai9           ; continue
        tst.l   d0
        bne.s   agai10          ; no extended info -> ..
        move.l  ioi_free(sp),d1 ; number of free sectors
        mulu    ioi_allc(sp),d1 ; * allocation size = free space
        tst.b   ioi_rdon(sp)    ; read only ??
        beq.s   agai11          ; no -> ..
        moveq   #err.rdo,d0     ; yes -> error
        bra.s   agai11
        ; oops, no extended information
agai10  move.l  sp,a1           ; place for medium name
        moveq   #iof.minf,d0    ; get medium information
        trap    #do.io
        cmpi.l  #err.nc,d0      ; not complete ?
        beq.s   agai10          ; continue
        clr.w   d1              ; throw out medium size
        swap d1                 ; put unused sectors in low word
        lsl.l   #8,d1
        add.l   d1,d1           ; fre space

agai11
        subend
;
        end
